;************************************************
;* ISO 2 MAGNETIC CARD ANALYZER [Mark II]
;*
;* Chip Number 		: U1
;* File Name 		: Program 013_2 Card Reader.asm
;* Date				: 06.20.2005
;* Last Update		: 06.20.2005
;* Version			: 2.0.0
;* Support Telephone: unlisted
;* Support Fax		: unlisted
;* Support E-mail	: brm016000@utdallas.edu
;* Target MCU		: AT90S2313-10PC
;*
;* DESCRIPTION
;*
;* Magnetic Card data content analyzer.  The following
;* program will read and store data recovered
;* from any magnetic data card (credit, debit, ID, etc) to 
;* the SRAM buffer.  Characters are displayed to 16x2 character
;* (LCD) display.  A red/green bi-color LED will display the
;* state of the system (orange="starting up", green="reading,:
;* red="parity error").  Two scrolling switches will let you
;* scroll through address 00d through 40d of the card.
;*
;* This program was intended for use with the Panasonic
;* ZU-M2121S451 Track 2 Reader.
;* 
;* This program was intended for use with a HD44780-compatible
;* 16x2 character LCD display.
;*
;* **NOTICE**
;*
;* Due to card reader limits the LRC checksum cannot be read
;* using this program.  This is due to the physical design of 
;* the Panasonic reader.
;*
;* PROGRAM CONCEPT
;* 
;* Scanning a single card:
;*
;* 1. Wakeup when "CARD DETECTED" or "CLOCK" pins go low 
;* 2. Hold until a falling CLOCK edge is detected
;* 3. Record value present at "DATA" pin and complement it.
;* 4. Add bit to 40-byte SRAM queue (starting at address 0x60)
;* 5. Continue steps 2 and 3 until 40 chars have been collected
;* 6. Display characters to the display while card is removed
;* 7. Reset cache when another card is inserted
;* 
;* PRESUMED CARD FORMAT
;* 
;* Characters will arrive as 5-bit words with the LSB first.
;* The logic is inverted.  The lowest four bits contain the
;* number or symbol.  Bit 5 is the parity bit, and is "set" 
;* when the lower four bits have an odd number of 1's.  Track 2
;* has 40 total characters; or a total of 200 bits.  HOWEVER:
;* THE READER'S END STOP WILL PHYSICALLY PREVENT ALL 200 BITS
;* FROM BEING READ.  The maximum data rate is assumed 1,772bps.
;*
;* To decode the data to ASCII, mask off the 5th (parity) bit
;* and add the value 0x30 to the remaining character bits.
;* 
;* Known Card Symbology:
;* [Start Marker]		- character 0x0B or ";"  One byte.
;* [End Marker]			- character 0x0F or "?"  One byte.
;* [Seperator]			- character 0x0D or "="  One byte.
;* [Account Number]		- UP TO 19 characters.   
;* [Country Code]		- present IF account starts with "59"  USA = 840. 3 bytes if used
;* [Expiration Date]	- YYMM format or 1 seperator if not present
;* [LRC checksum]		- information below
;*
;* TRACK 2 LAYOUT
;* <start sentinel>					1 byte
;* <primary account number> 		19 bytes maximum
;* <seperator>						1 byte
;* <country code>					3 bytes or 0 bytes
;* <expiration or serperator>		4 bytes or 1 byte
;* <discretionary data>				enough to fill out maximum record lenght
;* <longitudanal redundancy check>	1 byte
;*
;* Longitudanal Redundancy Check
;* 
;* 1 byte in length.  This final byte makes the total number of 1's in a 'row' of bits
;* even.  For example,  The LRC byte for the bytes 00001 10101 11111 00000 would be:
;* "01011".  
;*
;* Look at the first bit of all 4 bytes and add the 0's and 1's.  If the sum is 1,
;* then the first bit in the LRC byte should be 1.  If the sum is 0, then the first
;* LRC bit should be 0.  Continue this with the 2nd, 3rd, 4th, and 5th bits.
;*
;* PORT USAGE:
;* PORT D -- Card Reader Interface
;* [0] CARD DATA			input	(active low)
;* [1] CARD ENDSTOP			input	(active low)
;* [2] CARD DETECTED (INT0)	input	(active low)
;* [3] CLOCK (INT1)			input	(active low)
;* [4] Button 1				input	"Scroll Left"
;* [5] Button 2				input	"Scroll Right"
;* [6] -nc-
;*
;* PORT B -- Character Display Interface
;* [0] LCD data (lsb)		output
;* [1] LCD data				output
;* [2] LCD data				output
;* [3] LCD data (msb)		output
;* [4] LCD ENABLE			output	(active low)
;* [5] LCD REGISTER SELECT	output	(0=command 1=data)
;* [6] LED green			output
;* [7] LED red				output
;*************************************************	
;.device AT90S2313

.nolist
.include "2313def.inc"
.list

;- Register Aliases -----
.def	checksum		= R1	; LRC checksum
.def 	ones			= R2	; Used by d_deci routine
.def	tens			= R3	; Used by d_deci routine
;.def	accum			= R16	; General purpose register by tradition
.def	loop			= R17	; General loop counter by tradition
.def	addr			= R18	; General address holder by tradition
.def	addr2			= R19	; Secondary address holder
.def	parity			= R21	; Used for parity check shifting
.def	char_loop		= R22	; Character Count
.def	bit_loop		= R23	; Bit Count
.def	character		= R24	; 5-bit magstripe character
;- Define Constants -----
.equ	char_length		= 5		; Card character length (bits)
.equ	total			= 40	; Total Number of characters
.equ	EN				= 4
.equ	RS				= 5
.equ	pind_shadow		= PIND + $20
.equ 	data_mask		= 0b00000001 ; Mask all pins except PD.3
.equ 	card_capacity	= 40	; Total 5-bit characters
;- Macros ---------------
.macro RED
	sbi PORTB,6		; LRC "error" indicator
.endmacro

.macro GREEN
	sbi PORTB,7		; The "busy" indicator
.endmacro

.macro CLEAR
	cbi PORTB,6		; Turn off both indicators
	cbi PORTB,7
.endmacro

.macro lcd_standby
	sbi PORTB,EN	; Hold for instruction
.endmacro

.macro lcd_execute
	cbi PORTB,EN	; Run instruction on AVR port
.endmacro

.macro lcd_data
	sbi PORTB,RS	; Prepare LCD to receive characters
.endmacro

.macro lcd_command
	cbi PORTB,RS	; Prepare LCD for system command
.endmacro
;------------------------
.cseg						; Begin flash memory contents...
.org 0x00					; Begin at reset vector

	rjmp main	; Reset vector
;########################
;# Interrupt Vectors	#
;########################
	RETI		;Int0 Interrupt
	RETI		;Int1 Interrupt
	RETI		;TC1-Capture
	RETI		;TC1-Compare Match
	RETI		;TC1-Overflow
	RETI		;TC0-Overflow
	RETI		;UART Receive Transfer Complete
	RETI		;UART Data Register Empty
	RETI		;UART Transmit complete
	RETI		;Analog Comparator
;########################
;# End Interrupt Vectors#
;########################
main:	
	ser R16					; :::PORT B DIRECTION SETUP
	out DDRB,R16			; R16 = $FF : All pins are outputs

	clr R16					; ::: PORT B INITIALIZATION
 	out PORTB,R16			; All pins low until card activity

	clr R16					; :::PORT D DIRECTION SETUP
	out DDRD,R16			; R16 = $FF : All pins are inputs
	
	ser R16					; :::PORT D INITIALIZATION
	out PORTD,R16			; Engage internal pull-up resistors

	RED						; Create orange color to indicate...
	GREEN					; ..."initializing" stage

	ldi R16,(RAMEND)		; :::STACK SETUP INITIALIZATION
	out SPL,R16				; End-of-RAM address: 0xDF 

	ldi R16,0b00000100		; ::: TIMER-COUNTER ENABLE : Period = 64us
	out TCCR0,R16			; Send configuration

	ldi R16,0b00000010		; Timer interrupt enable mask
	out TIMSK,R16			; Send configuration

	ldi R16,0b11000000		; ::: EXTERNAL INTERRUPT ENABLE
	out GIMSK,R16			; Bit 6: INT0 (IRQ) interrupt ready
							; Bit 7: INT1 (IRQ) interrupt ready

	ldi R26,pind_shadow		; ::: X POINTER SETUP Load PORT D shadow register address
	ldi R27,0				; Clear upper byte

	ldi R28,0x60			; ::: Y POINTER RESET
	ldi R29,0x00			; 0x60 is first usable SRAM location

	rcall lcd_pre_init		; Pre-initialize LCD for 4-bit mode
	rcall lcd_setup			; Set LCD Visual Preferences

	ldi addr,0x02			; Move LCD cursor to center the text
	rcall d_cursor_pos		; Update cursor register in the HH44780
	ldi addr,msg_splash		; Load pointer to "DigiTan" in EEPROM
	rcall d_zt_str			; Display Zero-Terminated String

	ldi addr,0x40			; Move cursor to lower portion of screen
	rcall d_cursor_pos		; Update cursor register in the HH44780
	ldi addr,msg_splash2	; Load pointer to "Card Diagnostics"
	rcall d_zt_str			; Display Zero-Terminated String

	CLEAR					; Initialization done, turn off LEDs

	ldi R16,0b00111010		; ::: EXTERNAL INTERRUPT AND SLEEP ENABLE
	out MCUCR,R16			; Enable Sleep, Full Power Down,
							; Wakeup on falling INT1 or INT0
	
	sei						; Global Interrupt Enable
;	sleep					; Awake when card detected

	clr R16					; Disable Sleep mode
	out MCUCR,R16			; Prevent accidental sleep triggers

new_card:
	clr R16					; Disable external interrupts (IRQ)
	out GIMSK,R16			; INT0 and INT1 disabled 
;---CARD DETECTED!  BEGIN RECEIVING
	CLEAR
	GREEN					; Turn on "reading" status indicator

	ldi R28,0x60			; ::: Y POINTER RESET
	ldi R29,0x00			; 0x60 is first usable SRAM location

	rcall sram_clear		; Clear residual SRAM data

;---CARD DATA WILL BEGIN ARRIVING SOON, LISTEN FOR SERIAL DATA.
Card_receive:
	ldi char_loop,card_capacity	; Normal capacity is 40 characters
next_character:
	clr character			; Clear current character
	ldi bit_loop,char_length; Normal length is 5 bits
next_bit:
	lds R16,pind_shadow		; Read Port D status from SRAM
	andi R16,0b00001000		; Mask all pins except CLOCK
	breq next_bit			; Wait for previous bit to end
clock_listen:
	lds R16,pind_shadow		; Read Port D status from SRAM
	andi R16,0b00001000		; Mask all pins except CLOCK
	brne clock_listen		; Wait for new bit to arrive

	lds R16,pind_shadow		; Read Port D status from SRAM
	andi R16,data_mask		; Mask all pins except DATA bit
	brne received_NZ		; Branch if pin was a 5 volts
received_Z:
	sec						; *Set* Carry (bits are inverted)
	rjmp received_done
received_NZ:
	clc 					; *Clear* Carry (bits are inverted)
received_done:
	ror character			; Right-shift latest bit into register
	dec bit_loop			; Decrement loop counter
	brne next_bit			; Receive until character is complete

	lsr character			; Shift	right until lower 5 bits...
	lsr character			; ...of "character" register are...
	lsr character			; ...filled.
	st Y,character			; Store card data to SRAM
	inc R28					; Increment Y(SRAM) pointer
	dec char_loop			; Decrement character counter
	brne next_character		; Loop until 40 characters are collected

	CLEAR					; Turn off green "reading" LED

	rcall lcd_clear			; Clear away splash screen

	ldi addr,0x04			; PRINT "REMOVE"
	rcall d_cursor_pos		; Send LCD cursor to position 0-4
	ldi addr,msg_remove		; Point to "*Remove" in EEPROM
	rcall d_zt_str			; Display Zero-Terminate String

;---CARD HAS BEEN READ, HAVE USER REMOVE CARD
remove_wait:
	lds R16,pind_shadow		; Load PortD status byte
	andi R16,0b00000100		; Mask all but CARD DETECT
	breq remove_wait		; Hold until card is removed

	rcall lcd_clear			; Clear away word "*Remove*"

;--- DISPLAY CARD DIAGNOSTICS SCREEN
	ldi addr,0x00			; PRINT LEFT ARROW
	rcall d_cursor_pos		; Move LCD's cursor position
	ldi addr,msg_left		; Point to " <- " symbol in EEPROM
	rcall d_zt_str			; Display Zero-terminated String

	ldi addr,0x04			; PRINT "TRACK ISO2"	
	rcall d_cursor_pos		; Move LCD's cursor position
	ldi addr,msg_track2		; Point to string in EEPROM
	rcall d_zt_str			; Display Zero-terminated String

	ldi addr,0x0F			; PRINT RIGHT ARROW
	rcall d_cursor_pos		; Move LCD's cursor position
	ldi addr,msg_right		; Point to " -> " symbol in EEPROM
	rcall d_zt_str			; Display Zero-terminated String

	ldi addr,0x40			; PRINT CARD DATA OFFSET/ADDRESS
	rcall d_cursor_pos		; Move LCD's cursor position
	ldi addr,msg_bracket	; Display "[00]"
	rcall d_zt_str			; Display Zero-terminated String

	ldi R28,0x60			; Y-POINTER SETUP
	ldi R29,0x00			; Clear upper byte

show_contents:
	push R28				; Save Y-pointer start location
magstripe:
	rcall d_offset
	ldi loop,12				; Clear loop counter
	push loop				; Save
	ldi addr,0x44			; Start writing characters at this position
	rcall d_cursor_pos		; Starting cursor position
	pop loop				; Restore
display_data:
	push loop				; Re-save
	rcall parity_test		; OUT: If parity is OKAY...
							; addr = character + '0'   ASCII
							; OUT: If parity is BAD...
							; addr = 0xFF (ASCII for a black box)
	rcall d_char			; IN: addr from parity_test routine
	inc R28					; Increment target address
	pop loop				; Restore
	dec loop				; Increment loop counter	
	tst loop				; Is loop counter zero yet?
	brne display_data		; If not, draw more characters
	pop R28					; Restore SRAM(Y) Pointer start address

	ldi loop,24				; Lower data scroll rate to about 2 shifts/second
content_delay:
	push loop				; Save
	ldi loop,255			; Hold for 64us * 255
	rcall delay				; Maximum delay 16ms
	pop loop				; Restore
	dec loop				; Decrement
	brne content_delay		; Loop until zero

	ldi R26,pind_shadow		; ::: X POINTER SETUP Load PORT D shadow register address
	ldi R27,0				; Clear upper byte
keyloop:
	ld R16,X				; Load PORTD input data
	sbrs R16,4				; Skip unless "left" was pressed
	rjmp left
	sbrs R16,5				; Skip unless "right" was pressed
	rjmp right
	sbrs R16,2				; Allow re-scan if new card detected
	rjmp new_card
	rjmp keyloop

left:
	CLEAR
	dec R28					; Y-POINTER decrement
	cpi R28,0x60			; Reached end of SRAM queue?
	brne show_contents		; If not, update display
	ldi R28,0x88 - 0x0C		; If so, wrap to end of queue
	rjmp show_contents
right:
	CLEAR
	inc R28					; Y-POINTER increment
	cpi R28,0x88 - 0x0C		; Reaching end of SRAM queue?
	brne show_contents		; If not, update display
	ldi R28,0x60			; If so, wrap to start of queue
	rjmp show_contents

;---------------------------


lcd_pre_init:
	ldi loop,255		; Delay for 255*64us = 16.320ms (15ms minimum required)
	rcall delay			; Insert delay
	ldi R16,0b00010011	; Load $03 for 4-bit LCD interfacing.
	out PORTB,R16		; Out to LCD.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,72			; Delay of 72*64us = 4.608ms (4.1 minimum required)
	rcall delay			; Insert delay.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,2			; Delay of 2*64us (100us minimum required)
	rcall delay			; Insert delay.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,2			; Delay of 128us
	rcall delay			; Insert delay
	ldi R16,0b00010010	; Function set for 4-bit interface
	out PORTB,R16		; Out to LCD.
	LCD_EXECUTE
	LCD_STANDBY	
	ldi loop,2			; Delay of 128us.
	rcall delay			; Insert delay.
	ret					; End of pre-initialization.

lcd_setup:
	ldi R16,0b00010010	; FUNCTION SET :::: (BUS AND DISPLAY/FONT CONTROL) :::::
	out PORTB,R16		; -Data for 4-bit interface
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine
	ldi R16,0b00011000	; -Dual Line Display
	out PORTB,R16		; Send 2nd half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine.

	ldi R16,0b00010000	; DISPLAY OFF ::::: (DISPLAY POWER,CURSOR CONTROL)
	out PORTB,R16		; Send 1st half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine.
	ldi R16,0b00011000	; -Display off. -Cursors off. -Blinking off.
	out PORTB,R16		; Send 2nd half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us
	rcall delay			; Run delay routine.

	rcall lcd_clear		; DISPLAY CLEAR ::::: (DISPLAY CLEAR,CURSOR RETURN)

	ldi R16,0b00010000	; ENTRY MODE SETUP :::: (CURSOR AND SHIFTING CONTROL)
	out PORTB,R16		; Send 1st half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine.
	ldi R16,0b00010110	; - Cursor increment. - No display shifting
	out PORTB,R16		; Send 2nd half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine.

	ldi R16,0b00010000	; DISPLAY ON ::::: (DISPLAY POWER, CURSOR CONTROL)
	out PORTB,R16		; Send 1st half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us
	rcall delay			; Run delay routine.
	ldi R16,0b00011111	; -Display On. -Activate Cursor. -Activate Blinking.
	out PORTB,R16		; Send 2nd half of instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Run delay routine before returning.
	ret					; End of setup
;--------------------------------------------
delay:										;
; in: R17 = minimum delay = 64us * R17		;
; out: nothing								;
; destroyed: R16,R17,R6						;						;
;--------------------------------------------
	in R16,TCNT0		; Read current "time."
delay_sync:
	mov R6,R16			; Copy timer content to R6.
	in R16,TCNT0		; Read newer time.
	cp R16,R6			; Wait for a timer increment.
	breq delay_sync		; Loop until an increment occours first.
delay_start:
	in R16,TCNT0		; Get latest time.
delay_loop_0:
	mov R6,R16			; Copy timer content to R6.
	in R16,TCNT0		; Read newer time.
	cp R16,R6			; Wait for a timer increment.
	breq delay_loop_0	; Loop until in increment is detected
	dec R17				; Decrement LOOP counter.
	brne delay_start	; Run another 128us delay.
	ret					; Return.
;--------------------------------------------
lcd_clear:									;
; in: nothing
; out: nothing
; destroyed: R16,R17
;--------------------------------------------
	in R16,PORTB		; Get current PORTB configuration.
	andi R16,0b11110000	; Mask off control bits.
	out PORTB,R16		; Send upper half of CLEAR SCREEN instruction.
	LCD_COMMAND
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,28			; Pause for 28*64us = 1.792ms (1.64ms required).
	rcall delay			; Insert delay for execution.
	in R16,PORTB		; Get current PORTB configuration.
	andi R16,0b11110000	; Mask off control bits, as usual.
	ori R16,0b00000001	; Combine with CLEAR SCREEN instruction.
	out PORTB,R16		; Output to data bus.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,28			; Pause for 1.792ms.
	rcall delay			; Insert delay for execution.
	ret
;--------------------------------------------
d_char:
; in: R18 = ASCII character to load.
; out: nothing.
; destroyed: R16,R17,R18
;--------------------------------------------
	LCD_DATA
	push R18			; Save requested character
	in R16,PORTB		; Read current port configuration.
	andi R16,0b11110000	; Mask off control bits.
	andi R18,0b11110000	; Send upper half of character first.
	swap R18			; Swap nibbles to prepare for transmission.
	or R18,R16			; Combine bytes.
	out PORTB,R18		; Send data to bus.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Delay for 64us.
	rcall delay			; Insert delay.
	pop R18				; Restore original character.
	in R16,PORTB		; Restore R16
	andi R16,0b11110000	; Mask off control bits.
	andi R18,0b00001111	; Send lower half of character next.
	or R18,R16			; Combine bytes.
	out PORTB,R18		; Send data to bus.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Delay for 64us.
	rcall delay			; Insert delay.
	ret

d_offset:
	mov R18,R28			; Load SRAM pointer's offset
	subi R18,0x60		; Treat start-of-ram as 0-offset
	ldi R16,'0'			; Start at ASCII's "zero" character
	mov ones,R16		; Pre-set one's place counter
	mov tens,R16		; Pre-set ten's place counter
d_off_loop0:
	cpi R18,10			; BCD check
	brcs tens_done		; Branch if no tens remaining
	inc tens			; Otherwise, raise ten's counter
	subi R18,10			; Sub to check next ten's digit
	rjmp d_off_loop0
tens_done:
	add ones,R18		; Add directly to one's place
	ldi R18,0x41		; Move LCD text cursor to ten's place
	rcall d_cursor_pos	; "0x41" is "bottom row, column 1"
	mov R18,tens
	rcall d_char		; Show ten's digit
	mov R18,ones
	rcall d_char		; Show one's digit
	ret
;--------------------------------------------
d_zt_str:									;
; DISPLAY ZERO-TERMINATED CHARACTER STRING	;
; in: R18 = address of string
; out: nothing
; destroyed: R18,R16,R17
;--------------------------------------------
	out EEAR,R18		; Load target address.
	sbi EECR,EERE		; Enable EEPROM read.
	in R16,EEDR			; Load from data register.
	tst R16				; Check for zero. ("end-of-line")
	breq d_zt_escape	; Exit routine if zero is reached.
	LCD_DATA
	andi R16,0b11110000	; Send high-order bits first
	swap R16			; Send text data to lower nibble.
	push R18			; Save R18 content.
	in R18,PORTB		; Load current Port B status.
	andi R18,0b11110000	; Mask off data bits. Leave behind control and misc. bits.
	or R16,R18			; Combine bytes.
	out PORTB,R16		; Send to data bus.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Delay for 64us.
	rcall delay			; Wait for instruction to finish.
	in R16,EEDR			; Read once more from EEPROM data register.
	andi R16,0b00001111	; Now send lower-order bits.
	in R18,PORTB		; Load current Port B status.
	andi R18,0b11110000	; Mask off data bits. Leave behind control and misc. bits.
	or R16,R18			; Combine bytes.
	out PORTB,R16		; Send to data bus
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Delay for 64us.
	rcall delay			; Wait for full instruction to finish.
	pop R18				; Restore R18 content.
	inc R18				; Move to next EEPROM character.
	rjmp d_zt_str		; Loop again to display more characters.
d_zt_escape:
	ret					; Return after encountering a zero.

;--------------------------------------------
d_cursor_pos:								;
; ADJUST CURSOR POSITION					;
; in: R18 = position address (see descript.);
; out: nothing.
; destroyed: R16,R17,R18
;--------------------------------------------
	LCD_COMMAND
	ldi R16,$80			; Instruction for cursor positioning
	add R18,R16
	push R18			; Save R18's contents.
	andi R18,0b11110000	; Send upper command first.
	swap R18			; Move position data to lower nibble.
	in R16,PORTB		; Get current PORTB setting.
	andi R16,0b11110000	; Mask off control and misc. bits.
	or R16,R18			; Combine bytes.
	out PORTB,R16		; Output upper instruction
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Insert execution pause.
	pop R18				; Restore original R18
	andi R18,0b00001111	; Send lower command portion now.
	in R16,PORTB		; Get current PORTB setting again.
	andi R16,0b11110000	; Mask off control bits.
	or R16,R18			; Combine bytes.
	out PORTB,R16		; Output lower instruction.
	LCD_EXECUTE
	LCD_STANDBY
	ldi loop,1			; Pause for 64us.
	rcall delay			; Insert execution pause.
	ret

;------------------------
sram_clear:
; Reset all bytes within user SRAM
; IN: nothing
; OUT: nothing
; DESTROYED: R16
;------------------------
	push R28			; Save current Y-POINTER
	ldi R28,0x60		; Start with the first of user RAM
	clr R16				; R16 xor R16
sram_clear_loop:
	st Y,R16			; Clear current byte
	inc R28				; Move to next byte
	cpi R28,0x88+0x01	; Was all of user SRAM cleared?  (The end is really $7F)
	brne sram_clear_loop; If not, continue writing zeros.
	pop R28				; Restore current Y-POINTER
	ret					; Return

;------------------------------------------------
parity_test:
; Tests parity bit of next character against true 
; parity.  If the check was OK, the character is 
; adjusted for the right ASCII value and send to
; register "addr" (R18).  If the check was BAD,
; the character is replaced with an ASCII black box
; (value is 255).
;
; IN: nothing 
;
; [If parity is OK...]
; OUT: addr = adjusted ASCII character
; [If parity is BAD...]
; OUT: addr = 0xFF
;
; DESTROYED: R16, R17, R18, R1
;------------------------------------------------
	ldi loop,4				; Bits to parity-test
	clr parity				; Reset 1's counter
	ld R16,Y				; Read SRAM character
	mov R1,R16				; Save port status in a temp register
	andi R16,0b00001111		; Mask off parity bit
parity_test_loop0:
	lsr	R16					; Shift to carry
	brcc parity_no_change	; Zero's do not inc parity counter
	inc parity				; Increment parity-counter
parity_no_change:
	dec loop				; Decrement bits remaining counter
	brne parity_test_loop0
	mov R16,R1				; Restore SRAM character
	lsr parity				; Carry Set if odd parity
	brcs true_parity_odd
true_parity_even:
	andi R16,0b00010000		; Mask all but parity bit, update flags
	breq parity_ok
	rjmp parity_error
true_parity_odd:
	andi R16,0b00010000		; Mask all but parity bit
	brne parity_ok
parity_error:
	CLEAR
	RED
parity_ok:
	mov addr,R1				; Restore SRAM character
	andi addr,0b00001111	; Mask off parity bit
	ldi R16,'0'				; Load ASCII value for "zero"
	add addr,R16			; Adjust output character to 0-9
	ret
	



;########################
;# EEPROM Memory		#
;########################
.eseg
.org 0x02
;    XXXXXXXXXXXXXXXX
;    <  TRACK ISO2  >
;	   -=DigiTan=-
;	 Card Diagnostics
msg_splash:
.db "-=DigiTan=-",0			; <-- For the splash screen.  Change
msg_splash2:				; ...to whatever you like
.db "Card Diagnostics",0,0	; <-- Also for the splash
msg_track2:
.db "TRACK ISO2",0,0		; <-- Used on diagnostics screen
msg_bracket:
.db "[00]",0,0				; <-- Use for card offset
msg_right:
.db 0b01111110,0			; <-- ASCII for an "arrow"
msg_left:
.db 0b01111111,0			; <-- ASCII for an "arrow"
msg_remove:
.db "*Remove*",0,0